/*************************************************************************
**  Sony Unilink Interface (Main Procedure)
**  by Michael Wolf
**
**  Released under GNU GENERAL PUBLIC LICENSE
**	See LICENSE.TXT for details
**
**	See README.TXT for serial command details
**
**  Contact: michael@mictronics.de
**  homepage: www.mictronics.de
**
**  Revision History
**
**  when         what  who	why
**
**  2003-08-05   2.01  MIC  complete code revision
**                          + added jumper check during runtime
**                          - bugfix in time update command
**  2003-09-06	 2.02  MIC  * changed watchdog routine
**                          + added serial commands to change playmodes
**	2004-01-05   2.03  MIC  + added support for LIST function in CD mode
**														(used in Custom File CDC to choose a disc
**														by name from a list shown by the LIST button)
**													* changed display update routine for play modes
**													- bugfix in BANK mode display
**													* improved unilink_parse routine, this increase
**														reply speed like hell ;)
**	2004-02-27   2.04  MIC	* changed value for "missed ping" counter from
**														5 to 10 to solve problems with lots of bus
**														members especially with bus multiplexer units
**  2005-02-03   2.05  MIC  - removed all deprecaded avr-libc macros to be
**                            combatible with avr-libc v1.2.1 and higher
**
**  Used develompent tools (download @ www.avrfreaks.net):
**  Programmers Notepad v2.0.5.32
**  WinAVR (GCC) 3.4.1
**  AvrStudio4 for simulating and debugging
**	Tab Width: 2
**
**  [           Legend:          ]
**  [ + Added feature            ]
**  [ * Improved/changed feature ]
**  [ - Bug fixed (I hope)       ]
**
**************************************************************************/
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include "config.h"
#include "main.h"
#include "unilink.h"
#include "unilink.c"
#include "usart.c"
#include "general.c"

// Main loop
int main(void)
{
	cli();
	#if defined(WATCHDOG)
	wdt_enable(WDTO_2S);					// Watchdog enable with 2 seconds timeout
	#endif
	#if defined(__AVR_ATmega8__) || defined(__AVR_ATmega8515__)
	GICR = 0x00;     							// disable external interupts
	#else
	GIMSK = 0x00; 		            // disable external interupts
	#endif

	#if defined(__AVR_AT90S8515__) || defined(__AVR_ATmega8515__)
    MCUCR &=~ _BV(SRE);           // disable external RAM
	PORTA = 0x00;		            	// PortA all low
	DDRA = 0x00;		            	// PortA all input
	#endif
	PORTB = 0x00;		            	// PortB all low
	DDRB = 0x00;		            	// PortB all input
  PORTC = 0x00;		              // PortC all low
	DDRC = 0x00;		            	// PortC all input
  PORTD = 0x00;	                // PortD all low

	#if defined(__AVR_ATmega8__)
		#if defined(LED_OUT)
		DDRD = (1<<d_LED1)|(1<<d_LED2)|(1<<d_LED3);	// set PortD LED output on ATmega8
		#elif defined(RELAIS_OUT)
		DDRD = (1<<d_RELAIS);												// set PortD relais output on ATmega8
		#else
		DDRD = (0<<d_RELAIS)|(0<<d_LED1)|(0<<d_LED2)|(0<<d_LED3);	// set PortD all input on ATmega8
		#endif	
	#else
	DDRD = 0x00;											// PORTD all input on AT...8515
	#endif

  TIMSK = 1<<TOIE1;	  	            // enable Timer1 overflow interrupt
  TCCR0 = 0x00;     	              // Stop Timer0
	TCCR1A = 0x00;
	TCNT1 = c_1s;											// Load Timer1 with 1s
  TCCR1B = c_Timer1_run;		        // Start Timer1

	unilink_field.track = 0x0E;				// init name field counter
	unilink_field.disc = 0x0E;

  usart_init();                   	// init UART
	if (bit_is_set(PINB, b_JUMPER)) 	// check state of mode select jumper
  {
    unilink_ownaddr = C_UNILINK_OWNADDR_MD;// if set use MD mode
		unilink_groupid = C_UNILINK_OWNADDR_MD;
		flags.interface_mode = true;
  }
  else
  {
    unilink_ownaddr = C_UNILINK_OWNADDR_CD;// else use CD mode
		unilink_groupid = C_UNILINK_OWNADDR_CD;
		flags.interface_mode = false;
  };
	flags.unilink_anyone = false;
  sei();                    	     	// enable interrupts
	TCNT0 = c_6ms;		     		       	// load Timer0
  TCCR0 = c_Timer0_run;		    			// start Timer0

	SPCR = c_RUN_SPI;
  DDRB = 0x40;		            			// MISO as output and low
  SPDR = 0x00;

  // wait for idle time between two packets, then start receiving
  while(bit_is_clear(TIFR, TOV0))                             // wait for MOSI high for minimum of 6ms
	{                                
    if (bit_is_clear(PINB, b_MOSI)) TCNT0 = c_6ms;		    		// load Timer0 if MOSI is not high
  };// wait until timer0 overflow
  TIFR |= _BV(TOV0);                                         	// clear timer0 overflow flag
  TCNT0 = c_6ms;		                                        	// load Timer0
  while(bit_is_clear(TIFR, TOV0))                             // wait for MOSI low for minimum of 6ms
  {
    if (bit_is_set(PINB, b_MOSI)) TCNT0 = c_6ms;		    			// load Timer0  if MOSI is not low
  };// wait until timer overflow
  TCNT0 = c_Timer0_stop;		    															// stop timer0

	// now we are going in end endless loop waiting for packets
	while(true)
  {
		do
		{
			// check if jumper setting has changed after last reset
			if ( !flags.interface_mode && bit_is_set(PINB, b_JUMPER) )
				asm volatile ("rjmp 0");					// reset if CD mode and jumper set
			else if ( flags.interface_mode && bit_is_clear(PINB, b_JUMPER) )
				asm volatile ("rjmp 0"); 					// reset if MD mode and jumper not set
			#if defined WATCHDOG
			if(flags.bus_sleep)
				wdt_reset();												// reset watchdog here only if bus is in sleep mode
			#endif
		} while(!flags.unilink_rx_compl);
		
		flags.unilink_rx_compl = false;				// clear RX complete flag, wait for new packet
		checksum_check();
		
    if ( flags.checksum_ok )								// do a parity check of received packet and proceed if OK
    {
			unilink_parse();
			#ifdef BUS_LOGGING
        flags.unilink_tx_log = false;			// disable Tx packet logging
        bus_logging();                      // send valid Rx packet via UART
			#endif
    }  // end of routine if checksum was ok

		if(usart_rx_index > 0)
			usart_rx_proc();

	}; // end of endless while loop
}; // end of main()


// Timer1 Overflow interrupt
SIGNAL(SIG_OVERFLOW1)
{
	flags.timer1_ovf = true;							// set overflow flag
	TCNT1 = c_1s;           							// restart Timer1
	#if defined(WATCHDOG)
	unilink_missed_pings++;								/* increment missed ping counter every second,
                                           the counter is reset on each ping request
														               from head unit */
	if(unilink_missed_pings < 10)
		wdt_reset();
	#endif
};// end of TIMER1 overflow interrupt
